import tkinter as tk
from typing import Any

from subsearch.data import GUI_DATA, __paths__
from subsearch.data.data_objects import FormattedMetadata
from subsearch.gui import set_theme, tkinter_utils
from subsearch.gui.tabs import dowload_tab, language_tab, search_tab, settings_tab
from subsearch.utils import file_manager, raw_config, raw_registry


class TabManager(tk.Frame):
    """
    A class used to manage the tabs of a parent widget.

    Args:
        parent: The parent widget.
        tabs (dict[str, Any]): A dictionary containing tabs as a key-value pair. Key is the name of the tab and value is the content of the tab.
        active_tab (str): The current active tab.

    Attributes:
        parent: The parent widget.
        tabs (dict[str, Any]): A dictionary containing tabs as a key-value pair. Key is the name of the tab and value is the content of the tab.
        buttons: A dictionary containing button widgets as a key-value pair. Key is the name of the associated tab and value is the corresponding button widget.
        active_tab (str): The current active tab.

    Methods:
        activate_tabs(): Place the content of the active tab on the parent widget.
        release_tab(event): Called when a tab button is released. Set the active tab and call 'activate_tabs()' method.
        press_tab(event): Called when a tab button is pressed. Bind '<ButtonRelease>' event to the button widget.
        deactivate_tabs(): Hide the content of all tabs except the active one.
        enter_tab(event): Called when mouse enters a tab button. Bind '<ButtonPress>' event to the button widget.
        leave_tab(event):  Called when mouse leaves a tab button. Unbind '<ButtonPress>' event from the button widget, and call 'activate_tabs()' and 'deactivate_tabs()' methods.
        get_btn(dict_, event_, equals=True): Helper method to retrieve the button widget and its key.
    """

    def __init__(self, parent, tabs: dict[str, Any], active_tab: str) -> None:
        tk.Frame.__init__(self, parent)
        self.configure(bg=GUI_DATA.colors.mid_grey_black, width=GUI_DATA.size.root_width, height=82)
        relx_value = 0.0
        btn_kwargs: dict[str, Any] = dict(
            master=self, width=54, height=54, bg=GUI_DATA.colors.mid_grey_black, highlightthickness=0
        )
        self.parent = parent
        self.tabs = tabs
        self.buttons = {}

        for tab_key in tabs.keys():
            self.buttons[tab_key] = tk.Canvas(**btn_kwargs)

        for btn_key, btn_widget in self.buttons.items():
            relx_value += 0.2
            btn_widget.place(relx=relx_value, rely=0.5, anchor="center")
            btn_widget.bind("<Enter>", self.enter_tab)
            btn_widget.bind("<Leave>", self.leave_tab)
            tkinter_utils.asset_tab(btn_widget, btn_key, "rest")

        tkinter_utils.set_default_grid_size(self)
        self.active_tab = active_tab
        self.activate_tabs()

    def activate_tabs(self) -> None:
        self.tabs[self.active_tab].place(x=GUI_DATA.pos.content_x, y=GUI_DATA.pos.content_y, anchor="center")
        tkinter_utils.asset_tab(self.buttons[self.active_tab], self.active_tab, "press")
        self.parent.title(f"Subsearch - {self.active_tab} tab")

    def release_tab(self, event) -> None:
        btn_key, _btn_widget = self.get_btn(self.buttons, event)
        self.active_tab = btn_key
        self.activate_tabs()
        self.deactivate_tabs()

    def press_tab(self, event) -> None:
        btn_key, btn_widget = self.get_btn(self.buttons, event, False)
        btn_widget.bind("<ButtonRelease>", self.release_tab)
        tkinter_utils.asset_tab(btn_widget, btn_key, "press", y=20)

    def deactivate_tabs(self) -> None:
        for btn_key, btn_widget in self.buttons.items():
            if self.active_tab == btn_key:
                continue
            self.tabs[btn_key].place(x=GUI_DATA.pos.content_hidden_x, y=GUI_DATA.pos.content_y, anchor="nw")
            tkinter_utils.asset_tab(btn_widget, btn_key, "rest")

    def enter_tab(self, event) -> None:
        _btn_key, btn_widget = self.get_btn(self.buttons, event)
        btn_widget.bind("<ButtonPress>", self.press_tab)

    def leave_tab(self, event) -> None:
        _btn_key, btn_value = self.get_btn(self.buttons, event)
        btn_value.unbind("<ButtonPress>")
        self.activate_tabs()
        self.deactivate_tabs()

    def get_btn(self, dict_, event_, equals=True):
        for btn_key, btn_widget in dict_.items():
            if event_.widget == btn_widget and equals:
                return btn_key, btn_widget
            if event_.widget == btn_widget and not equals:
                return btn_key, btn_widget


class TabLanguage(tk.Frame):
    def __init__(self, parent) -> None:
        tk.Frame.__init__(self, parent, width=GUI_DATA.size.root_width, height=GUI_DATA.size.root_height)
        self.configure(bg=GUI_DATA.colors.dark_grey)
        language_tab.SelectLanguage(self).pack(anchor="center", expand=True)


class TabSearch(tk.Frame):
    def __init__(self, parent) -> None:
        tk.Frame.__init__(self, parent, width=GUI_DATA.size.root_width, height=GUI_DATA.size.root_height)
        self.configure(bg=GUI_DATA.colors.dark_grey)
        search_tab.Providers(self).pack(anchor="center")
        tk.Frame(self, height=80, bg=GUI_DATA.colors.dark_grey).pack(anchor="center", expand=True)
        search_tab.SubtitleType(self).pack(anchor="center")
        search_tab.SearchThreshold(self).pack(anchor="center")
        search_tab.RenameBestMatch(self).pack(anchor="center")


class TabSettings(tk.Frame):
    def __init__(self, parent) -> None:
        tk.Frame.__init__(self, parent, width=GUI_DATA.size.root_width, height=GUI_DATA.size.root_height)
        self.configure(bg=GUI_DATA.colors.dark_grey)
        settings_tab.FileExtensions(self).pack(anchor="center")
        tk.Frame(self, height=80, bg=GUI_DATA.colors.dark_grey).pack(anchor="center", expand=True)
        settings_tab.ShowContextMenu(self).pack(anchor="center")
        settings_tab.ShowContextMenuIcon(self).pack(anchor="center")
        settings_tab.ShowDownloadWindow(self).pack(anchor="center")
        settings_tab.UseThreading(self).pack(anchor="center")
        settings_tab.LogToFile(self).pack(anchor="center")
        if file_manager.running_from_exe() is False:
            settings_tab.ShowTerminalOnSearch(self).pack(anchor="center")
        tk.Frame(self, height=20, bg=GUI_DATA.colors.dark_grey).pack(anchor="center", expand=True)
        settings_tab.CheckForUpdates(self).pack(anchor="center")


class TabDownload(tk.Frame):
    def __init__(self, parent, formatted_data: list[FormattedMetadata]) -> None:
        tk.Frame.__init__(self, parent, width=GUI_DATA.size.root_width, height=GUI_DATA.size.root_height)
        self.configure(bg=GUI_DATA.colors.dark_grey)
        dowload_tab.DownloadList(self, formatted_data).pack(anchor="center")


def open_tab(active_tab: str, **kwargs) -> None:
    """
    Opens a new tab depending on the active_tab argument passed in.

    Args:
        active_tab (str): A string representing which tab to activate.
        **kwargs: Arbitrary keyword arguments that should contain "formatted_data",
                  a list of FormattedMetadata.

    Returns:
        None: This function does not return anything, it manipulates the GUI instead.
    """
    try:
        formatted_data: list[FormattedMetadata] = kwargs["formatted_data"]
    except KeyError:
        formatted_data = None  # type: ignore
    root = initalize_root()
    set_theme("dark")
    tkinter_utils.set_custom_btn_styles()
    tabs = {
        "language": TabLanguage(root),
        "search": TabSearch(root),
        "settings": TabSettings(root),
        "download": TabDownload(root, formatted_data),
    }
    footer = TabManager(root, tabs, active_tab.lower())
    footer.place(y=GUI_DATA.size.root_height - 82)
    root.mainloop()


def initalize_root():
    """
    Initializes the Tkinter root window with the name `"Subsearch"`, sets its background color to `GUI_DATA.colors.dark_grey`,
    adds an icon bitmap from `__paths__.icon`, configures the window geometry using `tkinter_utils.WindowPosition.set` and makes the window non-resizable.
    If `raw_registry.registry_key_exists()` is `False` and `raw_config.get_config_key("context_menu")` is `True`,
    the default JSON configuration is set using `raw_config.set_default_json()` and a context menu is added to the registry using `raw_registry.add_context_menu()`.

    Returns:
      An instance of Tk() class representing the main window of Subsearch application.
    """
    if raw_registry.registry_key_exists() is False and raw_config.get_config_key("context_menu"):
        raw_config.set_default_json()
        raw_registry.add_context_menu()
    root = tk.Tk(className=f"Subsearch")
    root.configure(background=GUI_DATA.colors.dark_grey)
    root.iconbitmap(__paths__.icon)
    root.geometry(tkinter_utils.WindowPosition.set(root))
    root.resizable(False, False)
    return root



import sys
from itertools import product
from pathlib import Path
from typing import no_type_check

from subsearch.data.data_objects import AppPaths, FileData


def paths() -> AppPaths:
    """
    Returns an instance of the AppPaths class representing the path configuration for Subsearch application.

    Returns:
        AppPaths: An instance of AppPaths representing the path configuration for Subsearch application.
    """
    home = Path(__file__).resolve().parent.parent
    return AppPaths(
        home=home,
        data=Path(home) / "data",
        gui=Path(home) / "gui",
        providers=Path(home) / "providers",
        utils=Path(home) / "utils",
        icon=Path(home) / "gui" / "assets" / "icon" / "subsearch.ico",
        tabs=Path(home) / "gui" / "assets" / "tabs",
    )


@no_type_check
def video_file() -> FileData:
    """
    Returns an instance of the FileData class based on the file path provided in the command line arguments.

    The function receives no inputs.

    Returns:
      An instance of the FileData class containing the video file data. None if the file path is not valid or not one of the supported extensions (.avi, .mp4, .mkv, .mpg, .mpeg, .mov, .rm, .vob, .wmv, .flv, .3gp, .3g2, .swf, .mswmm)

    FileData class:
      name (str): the name of the video file without the extension
      extension (str): the extension of the video file (e.g. ".avi")
      file_path (Path): the absolute path to the video file
      directory (Path): the parent directory of the video file
      subs_directory (Path): the absolute path to a directory to save the subtitles (same as the parent directory for the video file)
      tmp_directory (Path): the absolute path to the local temporary files directory (".subsearch" folder within the parent directory of the video file)
    """
    exts = [
        ".avi",
        ".mp4",
        ".mkv",
        ".mpg",
        ".mpeg",
        ".mov",
        ".rm",
        ".vob",
        ".wmv",
        ".flv",
        ".3gp",
        ".3g2",
        ".swf",
        ".mswmm",
    ]
    file_exist = False
    for i in product(exts, sys.argv):
        if i[1].endswith(i[0]) and str(i[1])[i[1].rfind("\\") :].startswith("\\"):
            file_path = Path(i[1])
            directory = file_path.parent
            tmp_directory = Path(directory) / ".subsearch"
            subs_directory = Path(directory) / "subs"
            name = file_path.stem
            ext = file_path.suffix
            file_exist = True
            break

    if file_exist:
        return FileData(name, ext, file_path, directory, subs_directory, tmp_directory)


__video__ = video_file()
__paths__ = paths()



import tkinter as tk
from pathlib import Path
from tkinter import Label, StringVar, ttk

from subsearch.data import GUI_DATA, __paths__, __version__, __video__
from subsearch.utils import file_manager, raw_config, raw_registry

GWL_EXSTYLE = -20
WS_EX_APPWINDOW = 0x00040000
WS_EX_TOOLWINDOW = 0x00000080

DEFAULT_LABEL_CONFIG = dict(bg=GUI_DATA.colors.dark_grey, fg=GUI_DATA.colors.white_grey, font=GUI_DATA.fonts.cas8b)
DEFAULT_LABEL_GRID = dict(row=0, column=0, sticky="w", padx=2, pady=2)
DEFAULT_BTN_TOGGLE_GRID = dict(row=0, column=2, pady=2)


def get_tab_png(tab: str) -> Path:
    """
    Get the path of the PNG file for a given tab.

    Args:
        tab (str): The name of the tab.

    Returns:
        Path: The path of the PNG file for the specified tab.

    Raises:
        None.
    """
    return Path(__paths__.tabs) / tab


def calculate_btn_size(cls, width_=18, height_=2) -> tuple[int, int]:
    """
    Calculate the required size of a button based on its content.

    Args:
        cls: The class from which the method is being called.
        width_ (optional): The initial width of the button before calculation. Defaults to 18.
        height_ (optional): The initial height of the button before calculation. Defaults to 2.

    Returns:
        A tuple of integer values representing the calculated width and height of the button
    """

    generic_btn = tk.Button(cls, width=width_, height=height_)
    x, y = generic_btn.winfo_reqwidth(), generic_btn.winfo_reqheight()
    return x, y


def calculate_checkbtn_size(cls, width_=16) -> int:
    """
    Calculates the required width of a ttk Checkbutton widget in pixels.

    Args:
        cls (class): The parent class where the Checkbutton will be placed.
        width_ (int, optional): The desired width of the ttk check button. Defaults to 16.

    Returns:
        int: The required width in pixels of a ttk Checkbutton based on its current settings.
    """

    generic_checkbtn = ttk.Checkbutton(cls, width=width_)
    return generic_checkbtn.winfo_reqwidth()


def set_default_grid_size(cls, width_=18) -> None:
    """
    Set the default grid size for a tkinter GUI window.

    Args:
        cls (object): The class object.
        width_ (int, optional): The default width. Defaults to 18.

    Returns:
        None
    """
    btn_size = tk.Button(cls, width=width_, height=2)
    x, _y = btn_size.winfo_reqwidth(), btn_size.winfo_reqheight()
    col_count, row_count = cls.grid_size()
    for col in range(col_count):
        cls.grid_columnconfigure(col, minsize=x)

    for row in range(row_count):
        cls.grid_rowconfigure(row, minsize=0)


def asset_tab(cls, img, type, x=27, y=27) -> None:
    """
    Attach an image displayed as a tab onto the application window.

    Args:
        cls (class): The class representing the application.
        img (str): The name of the image file.
        type (str): The type of the image file.
        x (int, optional): The width of the image in pixels. Defaults to 27.
        y (int, optional): The height of the image in pixels. Defaults to 27.
    """
    path = get_tab_png(f"{img}_{type}.png")
    png = tk.PhotoImage(file=path)
    update_asset(cls, png, x, y)


def update_asset(cls, img, x, y) -> None:
    """
    Updates an asset in a tkinter canvas with the provided image.

    Args:
        cls: The canvas instance that houses the current asset.
        img: The image asset to replace the current asset within the canvas.
        x: The X coordinate for positioning of the new asset.
        y: The Y coordinate for positioning of the new asset.

    Returns:
        None
    """

    cls.delete("all")
    cls.create_image(x, y, image=img)
    cls.photoimage = img


def set_custom_btn_styles() -> None:
    """
    Sets custom button styles

    Args:
        None

    Returns:
        None
    """

    custom_style = ttk.Style()
    custom_style.configure("True.TButton", foreground=GUI_DATA.colors.green)
    custom_style.configure("False.TButton", foreground=GUI_DATA.colors.red)


class WindowPosition(tk.Frame):
    """
    A class that creates and manages the positioning of a tkinter frame.

    Args:
        parent: The parent widget.

    Attributes:
        None

    Methods:
        set(w=int, h=int, ws_value_offset=int, hs_value_offset=int, other=bool): Sets the window dimension
            in width and height. It also sets offsets by which to move the window horizontally (ws_value_offset)
            and vertically (hs_value_offset). If other=True, it allows changing other dimensions not typically
            touched (eg. height of frame/window title bar)

    Returns:
       None
    """

    def __init__(self, parent):
        """
        Constructor Method for WindowPosition Class

        Args:
            parent: The parent widget.
        """
        tk.Frame.__init__(self, parent)

    def set(
        self,
        w=GUI_DATA.size.root_width,
        h=GUI_DATA.size.root_height,
        ws_value_offset=0,
        hs_value_offset=0,
        other: bool = False,
    ):
        """
        Set the size of the current window/Frame

        Args:
             w: An optional integer for setting the width of the window/frame
             h: An optional integer for setting the height of the window/frame
             ws_value_offset: An optional integer for setting the horizontal offset of the window/frame
             hs_value_offset: An optional integer for setting the vertical offset of the window/frame
             other: An optional bool variable to allow changing other dimensions not typically touched
                    eg. height of frame/window title bar

        Returns:
              None
        """
        ws = self.winfo_screenwidth() + ws_value_offset
        hs = self.winfo_screenheight() + hs_value_offset
        x = int((ws / 2) - (w / 2))
        y = int((hs / 2) - (h / 2))
        value = f"{w}x{h}+{x}+{y}"
        if other:
            return w, h, x, y
        return value


class VarColorPicker:
    """
    A class that initializes a color picker for given string variables.

    Args:
        string_var: A tkinter StringVar object to represent value of the color.
        clabel: A tkinter Label object to represent the color label.
        is_pct: A boolean variable initialized to False indicating if configured level in percentage.

    Attributes:
        string_var: A tkinter StringVar object representing value of the color.
        clabel: A tkinter Label object representing the color label.
        is_pct: A boolean variable to determine percentage threshold.

    Methods:
        pick(): Method to determine and update the color of clabel based on the string_var value.
    """

    def __init__(self, string_var: StringVar, clabel: Label, is_pct: bool = False):
        self.string_var = string_var
        self.clabel = clabel
        self.is_pct = is_pct
        self.pick()

    def pick(self):  # string boolean
        if self.string_var.get() == "True":
            self.clabel.configure(fg=GUI_DATA.colors.green)
        elif self.string_var.get() == "False":
            self.clabel.configure(fg=GUI_DATA.colors.red)
        elif self.string_var.get() == "Both":
            self.clabel.configure(fg=GUI_DATA.colors.blue)
        elif self.string_var.get().startswith("Only"):
            self.clabel.configure(fg=GUI_DATA.colors.green)

        if self.is_pct:
            _pct = raw_config.get_config_key("percentage_threshold")
            if _pct in range(75, 101):
                self.clabel.configure(fg=GUI_DATA.colors.green)
            elif _pct in range(50, 75):
                self.clabel.configure(fg=GUI_DATA.colors.green_brown)
            elif _pct in range(25, 50):
                self.clabel.configure(fg=GUI_DATA.colors.red_brown)
            elif _pct in range(0, 25):
                self.clabel.configure(fg=GUI_DATA.colors.red)


class ToolTip(tk.Toplevel):
    """
    A toplevel widget that displays a message when the user hovers over a specified widget

    Args:
        parent (tkinter.Tk): The parent widget
        _widget (tkinter.Widget): The widget to attach the tooltip to
        *_text (str): The text to be displayed in the tooltip
        _background (str): The background color of the tooltip

    Methods:
        show(): Creates and displays a toplevel widget containing the text to be displayed in the tooltip
    """

    def __init__(self, parent, _widget, *_text, _background=GUI_DATA.colors.light_black):
        self.parent = parent
        self.widget = _widget
        self.text = _text
        self.background = _background

    def show(self):
        tk.Toplevel.__init__(self, self.parent)
        self.configure(background=GUI_DATA.colors.light_black)
        # remove the standard window titlebar from the tooltip
        self.overrideredirect(True)
        # unpack *args and put each /n on a new line
        lines = "\n".join(self.text)
        frame = tk.Frame(self, background=GUI_DATA.colors.light_black)
        label = tk.Label(
            frame,
            text=lines,
            background=self.background,
            foreground=GUI_DATA.colors.white_grey,
            justify="left",
        )
        # get size of the label to use later for positioning and sizing of the tooltip
        x, y = label.winfo_reqwidth(), label.winfo_reqheight()
        # set the size of the tooltip background to be 1px larger than the label
        frame.configure(width=x + 1, height=y + 1)

        widget_pos = self.widget.winfo_rootx()
        widget_width = self.widget.winfo_reqwidth()
        widget_center = round(widget_width / 2)
        btn_pos_middle = widget_pos + widget_center
        frame_width = frame.winfo_reqwidth()
        frame_center = round(frame_width / 2)
        center_tip_over_mouse = btn_pos_middle - frame_center
        # offset the frame 1px from edge of the tooltip corner
        frame.place(x=1, y=1, anchor="nw")
        label.place(x=0, y=0, anchor="nw")
        root_x = center_tip_over_mouse  # offset the tooltip by extra 4px so it doesn't overlap the widget
        root_y = self.widget.winfo_rooty() - self.widget.winfo_height() - 4
        # set position of the tooltip, size and add 2px around the tooltip for a 1px border
        self.geometry(f"{x+2}x{y+2}+{root_x}+{root_y}")

    def hide(self):
        self.destroy()


class ToggleableFrameButton(tk.Frame):
    """
    A button that toggles a configuration setting on or off.

    Args:
        parent (tk.Widget): The parent widget.
        setting_label (str): Name of the setting to be displayeed.
        config_key (str): Key in the configuration file where the state is stored.
        write_to_reg (bool, optional): Whether to also write the state to registry. Defaults to False.
        show_if_exe (bool, optional): Only show the button if the program is not running from an executable. Defaults to True.
    """

    def __init__(self, parent, setting_label: str, config_key: str, write_to_reg: bool = False, show_if_exe=True) -> None:
        tk.Frame.__init__(self, parent)
        self.configure(bg=GUI_DATA.colors.dark_grey)
        self.string_var = tk.StringVar()
        self.string_var.set(f"{raw_config.get_config_key(config_key)}")
        self.setting_name = setting_label
        self.config_key = config_key
        self.write_to_reg = write_to_reg
        self.show_if_exe = show_if_exe
        if show_if_exe is False and file_manager.running_from_exe():
            return None
        label = tk.Label(self, text=self.setting_name)
        label.configure(DEFAULT_LABEL_CONFIG)
        label.grid(DEFAULT_LABEL_GRID)
        btn_toggle = ttk.Button(
            self,
            textvariable=self.string_var,
            width=40,
            style=f"{self.string_var.get()}.TButton",
        )
        btn_toggle.grid(DEFAULT_BTN_TOGGLE_GRID, padx=8)
        btn_toggle.bind("<Enter>", self.enter_button)
        btn_toggle.bind("<Leave>", self.leave_button)
        set_default_grid_size(self)

    def enter_button(self, event) -> None:
        """
        Function called when the mouse enters the button area.

        Args:
            event: The tkinter event object.
        """
        btn = event.widget
        if btn["text"] == "True":
            btn.bind("<ButtonRelease-1>", self.button_set_false)
        if btn["text"] == "False":
            btn.bind("<ButtonRelease-1>", self.button_set_true)

    def leave_button(self, event) -> None:
        """
        Function called when the mouse leaves the button area.

        Args:
            event: The tkinter event object.
        """
        btn = event.widget

    def button_set_true(self, event) -> None:
        """
        Function called when the button is set to True.

        Args:
            event: The tkinter event object.
        """
        btn = event.widget
        self.string_var.set(f"True")
        btn["style"] = f"{self.string_var.get()}.TButton"
        raw_config.set_config_key_value(self.config_key, True)
        if self.write_to_reg:
            raw_registry.add_context_menu()
            raw_registry.write_all_valuex()
        self.enter_button(event)

    def button_set_false(self, event) -> None:
        """
        Function called when the button is set to False.

        Args:
            event: The tkinter event object.
        """
        btn = event.widget
        self.string_var.set(f"False")
        btn["style"] = f"{self.string_var.get()}.TButton"
        raw_config.set_config_key_value(self.config_key, False)
        if self.write_to_reg:
            raw_registry.remove_context_menu()
        self.enter_button(event)



import socket
import sys
import winreg
from pathlib import Path

from subsearch.data import __paths__
from subsearch.utils import file_manager

COMPUTER_NAME = socket.gethostname()
CLASSES_PATH = "Software\\Classes"
ASTERISK_PATH = "Software\\Classes\\*"
SHELL_PATH = "Software\\Classes\\*\\shell"
SUBSEARCH_PATH = "Software\\Classes\\*\\shell\\Subsearch"
COMMAND_PATH = "Software\\Classes\\*\\shell\\Subsearch\\command"


def write_keys() -> None:
    """
    Writes keys to the windows registry.
    """
    key_paths = [(CLASSES_PATH, "*"), (ASTERISK_PATH, "shell"), (SHELL_PATH, "Subsearch"), (SUBSEARCH_PATH, "command")]
    with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_CURRENT_USER) as hkey:
        for path, sub_key in key_paths:
            with winreg.OpenKey(hkey, path, 0, winreg.KEY_WRITE) as sk:
                winreg.CreateKey(sk, sub_key)


def write_all_valuex() -> None:
    """
    Write values to the registry for `subsearch`, `icon`, `appliesto` and `command`.
    """
    write_valuex("subsearch")
    write_valuex("icon")
    write_valuex("appliesto")
    write_valuex("command")


def write_valuex(key: str) -> None:
    """
    Write a value to registry key.

    Args:
        key: A string that represents the target registry key.

    Returns:
        None.
    """
    if key == "subsearch":
        key_type = SUBSEARCH_PATH
        value_name = ""
        value = "Subsearch"
    if key.lower() == "icon":
        key_type = SUBSEARCH_PATH
        value_name = "Icon"
        value = get_icon_value()
    elif key.lower() == "appliesto":
        key_type = SUBSEARCH_PATH
        value_name = "AppliesTo"
        value = get_appliesto_value()
    elif key.lower() == "command":
        key_type = COMMAND_PATH
        value_name = ""
        value = get_command_value()
    open_write_valuex(key_type, value_name, value)


def open_write_valuex(sub_key: str, value_name: str, value: str) -> None:
    """
    Writes a value to a specific key in the Windows registry.

    Args:
        sub_key (str): The path of the registry key.
        value_name (str): The value name that the data will be written to.
        value (str): The data to be written to the registry.

    Raises:
        FileNotFoundError: If the specified registry key or value does not exist.
    """
    try:
        # connect to Computer_NAME\HKEY_CURRENT_USER\
        with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_CURRENT_USER) as hkey:
            # open key, with write permission
            with winreg.OpenKey(hkey, sub_key, 0, winreg.KEY_WRITE) as sk:
                # write valuex to matching value name
                winreg.SetValueEx(sk, value_name, 0, winreg.REG_SZ, value)
    except FileNotFoundError:
        pass


def get_command_value() -> str:
    """
    Returns the command to execute when the context menu is used on a file for SubSearch.

    Returns:
        A string containing the command to execute for the corresponding value.
    """
    # get latest json value from file
    from subsearch.utils import raw_config

    if file_manager.running_from_exe():
        value = f'"{sys.argv[0]}" "%1"'
        # if SubSearch is compiled we dont need anything besides this
    elif file_manager.running_from_exe() is False:
        show_terminal = raw_config.get_config_key("show_terminal")
        # gets the location to the python executable
        python_path = Path(sys.executable).parent
        # sys.args[-1] is going to be the path to the file we right clicked on
        # import_sys = "import sys; media_file_path = sys.argv[-1];"
        set_title = "import ctypes; ctypes.windll.kernel32.SetConsoleTitleW('subsearch');"
        # gets the path of the root directory of subsearch
        set_wd = f"import os; os.chdir(r'{__paths__.home}');"
        import_main = "import subsearch; subsearch.main()"
        if show_terminal is True:
            value = f'{python_path}\python.exe -c "{set_title} {set_wd} {import_main}" "%1"'
        if show_terminal is False:
            value = f'{python_path}\pythonw.exe -c "{set_title} {set_wd} {import_main}" "%1"'

    return value


def get_icon_value() -> str:
    """
    Returns the icon path to use in the context menu, or an empty string if configuration specifies it should not be shown

    Returns:
        str: Path of the icon file to be used in the context menu, or an empty string if the configuration specifies
             that it should not be shown.
    """
    from subsearch.utils import raw_config

    show_icon: str = raw_config.get_config_key("context_menu_icon")
    if show_icon:
        return str(__paths__.icon)
    else:
        return ""


def get_appliesto_value() -> str:
    """
    Retrieve the latest json value from file `raw_config`.

    Returns:
        str: A string of file types to show the SubSearch context entry on
            The file types are concatenated with `" OR "` in between.
    """
    # get latest json value from file
    from subsearch.utils import raw_config

    file_ext = raw_config.get_config_key("file_extensions")
    # for which file types to show the SubSearch context entry on
    value = ""
    for k, v in zip(file_ext.keys(), file_ext.values()):
        if v is True:
            value += "".join(f'"{k}" OR ')
    if value.endswith(" OR "):  # remove last OR
        value = value[:-4]

    return value


def remove_context_menu() -> None:
    with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_CURRENT_USER) as hkey:
        with winreg.OpenKey(hkey, SUBSEARCH_PATH, 0, winreg.KEY_WRITE) as sk:
            winreg.DeleteKey(sk, "command")

    with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_CURRENT_USER) as hkey:
        with winreg.OpenKey(hkey, SHELL_PATH, 0, winreg.KEY_WRITE) as sk:
            winreg.DeleteKey(sk, "Subsearch")


# write keys to registry then write values
def add_context_menu() -> None:
    write_keys()
    write_all_valuex()


def registry_key_exists() -> bool:
    """
    Check if the Subsearch context menu key is already present in the Windows registry.

    Returns:
        bool: True, If the key exists in the registry, else False.
    """
    sub_key = rf"Software\Classes\*\shell\Subsearch\command"
    try:
        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as hkey:
            winreg.OpenKey(hkey, sub_key)
            return True
    except Exception:
        return False


def key_no_value() -> bool:
    """
    Checks if the "command" value name has no value.

    Returns:
        True if the "command" value name does not have a value, False otherwise.
    """
    sub_key = rf"Software\Classes\*\shell\Subsearch\command"
    with winreg.ConnectRegistry(COMPUTER_NAME, winreg.HKEY_CURRENT_USER) as hkey:
        with winreg.OpenKey(hkey, sub_key) as subkey:
            default = winreg.QueryValueEx(subkey, None)  # type: ignore
            if default[0] == "":
                return True
    return False



import re

from subsearch.providers import generic


class AdvTitleSearch:
    """
    A class representing an advanced title search.

    Attributes:
        title (str): The title of the movie or TV series.
        year (int): The release year of the movie or TV series.

    Methods:
        get_url(): Returns the URL for searching the given title and year on IMDb.

    Properties:
        prior_year: Returns a string with the start date of the previous year in ISO format.
        release_year: Returns a string with the end date of the current year in ISO format.
        release_title: Returns the title with spaces replaced by '+' signs for query purposes.
        title_search: Returns a string with the IMDB search parameter for the given title.
        release_date_search: Returns a string with the IMDB search parameter for the release date of the movie or TV series.
    """

    def __init__(self, title: str, year: int) -> None:
        """Initializes an instance of the AdvTitleSearch class."""
        self.title = title
        self.year = year

    def get_url(self) -> str:
        """Returns the URL for the advanced title search."""
        imdb_domain = "https://www.imdb.com"
        return f"{imdb_domain}/search/title/?{self.title_search}&{self.release_date_search}"

    @property
    def prior_year(self) -> str:
        """Returns a string with the start date of the previous year in ISO format."""
        return f"{self.year-1}-01-01"

    @property
    def release_year(self) -> str:
        """Returns a string with the end date of the current year in ISO format."""
        return f"{self.year}-12-31"

    @property
    def release_title(self) -> str:
        """Returns the title with spaces replaced by '+' signs for query purposes."""
        return self.title.replace(" ", "+")

    @property
    def title_search(self) -> str:
        """Returns a string with the IMDB search parameter for the given title."""
        return f"title={self.release_title}"

    @property
    def release_date_search(self) -> str:
        """Returns a string with the IMDB search parameter for the release date of the movie or TV series."""
        return f"release_date={self.prior_year},{self.release_year}"


class FindImdbID(AdvTitleSearch):
    """
    This class is used to find the IMDb ID of a movie or TV show based on its title and year.

    Args:
        title (str): The title of the movie or TV show.
        year (int): The year the movie or TV show was released.

    Attributes:
        title (str): The lowercased title of the movie or TV show.
        year (int): The year the movie or TV show was released.
        id (str or None): The IMDb ID of the movie or TV show if found. Otherwise, set to None.

    Methods:
        find_imdb_title: This method finds the title of a movie or TV show on IMDb.
        find_imdb_year: This method finds the year of release of a movie or TV show on IMDb.
        get_imdb_id: This method retrieves the IMDb ID of a movie or TV show from its page on IMDb.

    """

    def __init__(self, title: str, year: int):
        """Initializes FindImdbID with a title and year."""

        self.title = title.lower()
        self.year = year
        self.id = None

        adv_search = AdvTitleSearch(self.title, self.year)

        url = adv_search.get_url()
        tree = generic.get_html_parser(url)
        product = tree.select("div.lister-item-content")

        for item in product.matches:
            self.data = item.css_first("h3.lister-item-header")

            if self.title != self.find_imdb_title():
                continue

            if self.year != self.find_imdb_year() and (self.year - 1) != self.find_imdb_year():
                continue

            self.id = self.get_imdb_id()
            break

    def find_imdb_title(self) -> str:
        """
        Finds the title of a movie or TV show on IMDb.

        Returns:
            str: The title of the movie or TV show.
        """

        title = self.data.css_first("a").text()
        return title.lower()

    def find_imdb_year(self) -> int:
        """
        Finds the year of release of a movie or TV show on IMDb.

        Returns:
            int: The year the movie or TV show was released.
        """

        year = self.data.css_first("span.lister-item-year").child.text_content
        return int(re.findall("[0-9]+", year)[0])

    def get_imdb_id(self) -> str:
        """
        Retrieves the IMDb ID of a movie or TV show from its page on IMDb.

        Returns:
            str: The IMDb ID of the movie or TV show.
        """

        href = self.data.css_first("a")
        return re.findall("tt[0-9]+", href.attributes["href"])[0]



from dataclasses import dataclass
from pathlib import Path


@dataclass(order=True)
class AppPaths:
    """
    A dataclass representing the paths used in the application.

    Attributes:
        home (Path): The root directory of the application.
        data (Path): The directory where data related to the application is stored.
        gui (Path): The directory containing GUI modules and files.
        providers (Path): The directory containing subtitle provider modules.
        utils (Path): The directory containing utility modules.
        icon (Path): The Path object representing the icon for the application.
        tabs (Path): The directory containing tab icon assets for the GUI.
    """

    home: Path
    data: Path
    gui: Path
    providers: Path
    utils: Path
    icon: Path
    tabs: Path


@dataclass(order=True)
class FileData:
    """
     Class representing file information.

    Attributes:
        filename (str): Name of the file.
        file_extension (str): Extension of the file.
        file_path (pathlib.Path): Path of the file.
        directory_path (pathlib.Path): Directory path of the file.
        subs_directory (pathlib.Path): Subtitles directory path.
        tmp_directory (pathlib.Path): Temporary directory path.
    """

    filename: str
    file_extension: str
    file_path: Path
    directory_path: Path
    subs_directory: Path
    tmp_directory: Path


@dataclass(order=True)
class AppConfig:
    """
    Data class for storing user preferences and settings.

    Attributes:
        current_language (str): The currently selected language.
        languages (dict[str, str]): A mapping of languages.
        subtitle_type (dict[str, bool]): A dictionary used to store user's preference regarding subtitle types.
        percentage_threshold (int): Percentage threshold value for preferred match results.
        rename_best_match (bool): Boolean flag to rename the closest matching srt file with the media file used in search.
        context_menu (bool): Boolean flag representing whether or not to display SubSearch in the context menu on Windows platform.
        context_menu_icon (bool): Boolean flag representing whether or not to display the SubSearch icon in the context menu on Windows platform.
        manual_download_fail (bool): Boolean flag indicating if the GUI download tab should be opened.
        manual_download_mode (bool): Boolean flag to toggle manual download mode.
        show_terminal (bool): Boolean flag indicating if terminal output is displayed.
        use_threading (bool): Boolean flag indicating if threads are enabled when downloading subtitles.
        log_to_file (bool): Boolean flag that indicates if logging is enabled.
        file_extensions (dict[str, bool]): A dictionary storing user's preferences regarding file extensions.
        providers (dict[str, bool]): A dictionary storing user's preferences regarding subtitle providers.
        language_iso_639_3 (str): ISO 639-3 code for the selected language.
        hearing_impaired (bool): Boolean flag indicating if the subtitle is for hearing-impaired people.
        non_hearing_impaired (bool): Boolean flag indicating if the subtitle is not for hearing-impaired people.
    """

    current_language: str
    languages: dict[str, str]
    subtitle_type: dict[str, bool]
    percentage_threshold: int
    rename_best_match: bool
    context_menu: bool
    context_menu_icon: bool
    manual_download_fail: bool
    manual_download_mode: bool
    show_terminal: bool
    use_threading: bool
    log_to_file: bool
    file_extensions: dict[str, bool]
    providers: dict[str, bool]
    language_iso_639_3: str
    hearing_impaired: bool
    non_hearing_impaired: bool


@dataclass(frozen=True, order=True)
class ReleaseMetadata:
    """
    Data class representing metadata associated with a media release.

    Attributes:
        title (str): Title of the media.
        year (int): Year of the media release.
        season (str): Season of the media.
        season_ordinal (str): Ordinal of the season.
        episode (str): Episode of the media.
        episode_ordinal (str): Ordinal of the episode.
        tvseries (bool): Boolean flag indicating if the media is a TV series.
        release (str): Name of the media release.
        group (str): Name of the release group.
        file_hash (str): Hash value of the media file.
    """

    title: str
    year: int
    season: str
    season_ordinal: str
    episode: str
    episode_ordinal: str
    tvseries: bool
    release: str
    group: str
    file_hash: str


@dataclass(frozen=True, order=True)
class ProviderUrls:
    """
    A dataclass to represent URLs for different subtitle providers.

    Attributes:
        subscene(str): URL of the subscene provider.
        opensubtitles(str): URL of the opensubtitle provider.
        opensubtitles_hash(str): Hash URL of the opensubtitle provider.
        yifysubtitles(str): URL of the yifysubtitles provider.
    """

    subscene: str
    opensubtitles: str
    opensubtitles_hash: str
    yifysubtitles: str


@dataclass(frozen=True, order=True)
class DownloadMetaData:
    """
    A data class representing metadata for the to be downloaded subtitle file.

    Attributes:
        provider (str): The subtitle provider used to obtain the subtitle file.
        name (str): The name of the media file associated with the subtitle file.
        file_path (str): The path to the subtitle file on the local filesystem.
        url (str): The URL from which the subtitle file can be downloaded from.
        idx_num (int): Index in which order the file is downloaded.
        idx_length (int): The total length of index.
    """

    provider: str
    name: str
    file_path: str
    url: str
    idx_num: int
    idx_lenght: int


@dataclass(frozen=True, order=True)
class FormattedMetadata:
    """
    Represents formatted metadata for a movie or TV show release.

    Attributes:
        provider (str): The name of the metadata provider.
        release (str): The name of the movie or TV show release.
        url (str): The URL from which the subtitle file can be downloaded from.
        pct_result (int): The percentage match between the search query and
            the metadata.
        formatted_release (str): The formatted name of the release.
        formatted_url (str): The formatted URL of the metadata.
    """

    provider: str
    release: str
    url: str
    pct_result: int
    formatted_release: str
    formatted_url: str


@dataclass(frozen=True, order=True)
class GUISizes:
    """
    Sub dataclass containing sizes for the GUI
    """

    root_width: int = 738
    root_height: int = 720


@dataclass(frozen=True, order=True)
class GUIPositions:
    """
    Sub dataclass containing sizes for the GUI
    """

    content_x: int = int(GUISizes().root_width / 2)
    content_y: int = int(GUISizes().root_height / 2 - 41)
    content_hidden_x: int = int(GUISizes().root_width * 2)


@dataclass(frozen=True, order=True)
class GUIColors:
    """
    Sub dataclass containing colors for the GUI

    Colors:
        purple: #b294bb
        red: #bc473b
        dark_red: #89332a
        red_brown: #b26947,
        orange: #ab7149
        light_orange: #de935f
        yellow: #f0c674
        blue: #81a2be
        cyan: #82b3ac
        green: #9fa65d
        green_brown: #a59256
        grey: #4c4c4c
        light_grey: #727272
        silver_grey: #8a8a8a
        white_grey: #bdbdbd
        dark_grey: #1a1b1b
        mid_grey_black: #111111
        light_black: #0e0e0e
        black: #151515
        dark_black: #000000
    """

    purple: str = "#b294bb"
    red: str = "#bc473b"
    dark_red: str = "#89332a"
    red_brown: str = "#b26947"
    orange: str = "#ab7149"
    light_orange: str = "#de935f"
    yellow: str = "#f0c674"
    blue: str = "#81a2be"
    cyan: str = "#82b3ac"
    green: str = "#9fa65d"
    green_brown: str = "#a59256"
    grey: str = "#4c4c4c"
    light_grey: str = "#727272"
    silver_grey: str = "#8a8a8a"
    white_grey: str = "#bdbdbd"
    dark_grey: str = "#1a1b1b"
    mid_grey_black: str = "#111111"
    light_black: str = "#0e0e0e"
    black: str = "#151515"
    dark_black: str = "#000000"


@dataclass(frozen=True, order=True)
class GUIFonts:
    """
    Sub dataclass containing fonts and their styles and or size

    Fonts:
        cas6b: Cascadia 6 bold
        cas8: Cascadia 8
        cas8i: Cascadia 8 italic
        cas8b: Cascadia 8 bold
        cas10b: Cascadia 10 bold
        cas11: Cascadia 11
        cas20b: Cascadia 20 bold
    """

    cas6b: str = "Cascadia 6 bold"
    cas8: str = "Cascadia 8"
    cas8i: str = "Cascadia 8 italic"
    cas8b: str = "Cascadia 8 bold"
    cas10b: str = "Cascadia 10 bold"
    cas11: str = "Cascadia 11"
    cas20b: str = "Cascadia 20 bold"


@dataclass(frozen=True, order=True)
class GUIData:
    """
    Dataclass containing GUI data
    """

    size = GUISizes
    pos = GUIPositions
    fonts = GUIFonts
    colors = GUIColors


GUI_DATA = GUIData()



from importlib import util

# TODO: Modify the MSI installer to add the correct registry key values without relying on the application to run once to generate them.
# Currently, the installer only adds registry key placeholders, which are updated only after the app runs and generates the correct values
# . 

def _cx_freeze():
    """
    Builds an executable and MSI installer using cx_Freeze.

    """
    from cx_Freeze import Executable, setup
    icon = "src/subsearch/gui/assets/icon/subsearch.ico"
    # Define registry keys and entries for context menu integration in Windows Explorer
    #
    # The context menu will be added to all file types. This can be later customized
    # inside the application itself, as the applicability of the context menu is associated
    # with file extensions and controlled by the `appliesto` entry in the registry.
    #
    # For more information on the MSI registry table, see:
    # https://learn.microsoft.com/en-us/windows/win32/msi/registry-table
    app_name = "Subsearch"
    registry_path = rf"Software\Classes\*\shell\Subsearch"
    script_component = "_cx_executable0__Executable_script_src_subsearch___main__.py_"
    subsearch_key = (f"{app_name}_key", -1, registry_path, None, None, script_component)
    subsearch_icon = (f"{app_name}_regz_icon", -1, registry_path, "Icon", None, script_component)
    subsearch_appliesto = (f"{app_name}_regz_appliesto", -1, registry_path, "AppliesTo", None, script_component)
    subsearch_command = (f"{app_name}_key_command", -1, rf"{registry_path}\command", None, "", script_component)

    msi_upgrade_code = "{F12EBFBB-927E-4D29-818F-F26DB025D16C}"
    msi_data = {"Registry": {subsearch_key, subsearch_icon, subsearch_appliesto, subsearch_command}}

    bdist_msi_options = {"upgrade_code": f"{msi_upgrade_code}", "install_icon": icon, "data": msi_data}

    executable = [
        Executable(
            "src/subsearch/__main__.py",
            base="Win32GUI",
            target_name="Subsearch",
            icon=icon,
            shortcut_name="Subsearch",
            shortcut_dir="ProgramMenuFolder",
        ),
    ]
    setup(options={"bdist_msi": bdist_msi_options}, executables=executable)


def _setup_tools():
    """
    Builds the package using setuptools.

    """
    import setuptools

    setuptools.setup()


def setup():
    """
    Determines whether cx_Freeze is installed and calls the appropriate build function.

    """
    cx_freeze_installed = util.find_spec("cx_Freeze")
    if cx_freeze_installed:
        return _cx_freeze()
    return _setup_tools()


if __name__ == "__main__":
    setup()
